
#include <math.h>

#include <functional>
#include <algorithm>

#include "body.h"
#include "face.h"
#include "exception.h"
#include "model.h"

FeBody::FeBody() {
  mvCenterValid=false;

  mvDensity=0.0;

  mvVolume=0.0;
  mvPercolatedVolume=-1.0;
  mvWeight=0.0;  
}

FeBody::~FeBody() {
  DEBUG(10, cout << "\t\tDestroying body " << id() << endl);
}

FeModel *FeBody::model() {
  return mvpModel;
}

void FeBody::initialVelocity(FeVector &pv) {
  mvInitialVelocity=pv;
}

void FeBody::initialVelocity(double *pv) {
  FeVector nv(pv);
  initialVelocity(nv);
}

FeVector *FeBody::initialVelocity() {
  return &mvInitialVelocity;
}

void FeBody::velocity(FeVector &pv) {
  mvVelocity=pv;
  //  mvVelocity.normalize();
}

double FeBody::density(double d) {
  if (d != 0.0) mvDensity=d;

  if (mvDensity == 0.0) throw FeError("Body has zero density");

  return mvDensity;
}

FeVector FeBody::velocity() {
  return mvVelocity;
}

void FeBody::addFace(FeFace *pf) {
  mvpFaces.push_back(pf);

  pf->setBody(this);
}

vector<FeFace *>::iterator FeBody::firstFace() {
  return mvpFaces.begin();
}

vector<FeFace *>::iterator FeBody::lastFace() {
  return mvpFaces.end();  
}

void FeBody::checkTopology() {
}

struct Counter: public unary_function<FeFace *, void> {
  Counter(FeVector *resC, int *resNum) { mResC=resC; mResNum=resNum; };
  FeVector *mResC;
  int *mResNum;
  void operator () (FeFace *pf) {
    FeVector pc=pf->center();
    *mResC += pc;
    (*mResNum)++;
  };
};

int FeBody::number(int newNumber) {
  if (newNumber) mvNumber=newNumber;

  return mvNumber;
}

FeVector FeBody::center() {
  if (!mvCenterValid) {
    FeVector resC;
    int resNum=0;
    Counter pc1(&resC, &resNum);
    for_each(mvpFaces.begin(), mvpFaces.end(), pc1);
    
    mvCenter=resC/(double)resNum;
  }

  return mvCenter;
}

struct VolumeCounter: public unary_function<FeFace *, void> {
  VolumeCounter(double *res, FeVector &cen):mCen(cen) { mRes=res; };
  double *mRes;
  FeVector &mCen;
  void operator () (FeFace *pf) {
    double dist=fabs(pf->vertexDist(mCen));
    double area=pf->area();
    *mRes += dist*area/3.0;
  };
};

double FeBody::volume() {
  if (mvVolume == 0.0) {
    FeVector pc=center();
    double resVol=0.0;
    VolumeCounter pc1(&resVol, pc);
    for_each(mvpFaces.begin(), mvpFaces.end(), pc1);

    mvVolume=resVol;
  }

  return mvVolume;
}

double FeBody::percolatedVolume() {
  if (mvPercolatedVolume < 0.0) {
    mvPercolatedVolume=0.0;

    if (mvpModel->waterLevel() != FE_INFINITY) {
      FeVector pc=center();
      FeVector pcwl(pc[0], pc[1], mvpModel->waterLevel());
      
      for (vector<FeFace *>::iterator actF=mvpFaces.begin(); actF < mvpFaces.end(); actF++) {
	double pa=(*actF)->percolatedArea();
	double dist=(*actF)->vertexDist(pcwl);
	mvPercolatedVolume += -pa*dist/3.0;
      }
    }
  }

  return mvPercolatedVolume;
}

double FeBody::weight() {
  if (mvWeight == 0.0) {
    mvWeight=GRAVITY * volume() * density();
  }

  return mvWeight;
}

FeVector& FeBody::externalForces(FeVector &v) {
  return mvExternalForces=v;
}

FeVector& FeBody::externalForces() {
  return mvExternalForces;
}

FeVector& FeBody::equilibriumForceDirection(FeVector &v) {
  return mvEquilibriumForceDirection=v;
}

void FeBody::invalidatePrecomputedValues() {
  mvCenterValid=false;
  mvVolume=0.0;
  mvPercolatedVolume=-1.0;
  mvWeight=0.0;
}

FeVector& FeBody::equilibriumForceDirection() {
  return mvEquilibriumForceDirection;
}

void FeBody::calculateCohesionForces() {
  mvCohesionForcesSum=FeVector(0, 0, 0);
  
  vector<FeFace *>::iterator actF=mvpFaces.begin();
  while (actF != mvpFaces.end()) {
    FeVector actCF=(*actF)->cohesionForce();

    mvCohesionForces.push_back(actCF);
    mvCohesionForcesSum += actCF;

    actF++;
  }
}

FeVector& FeBody::cohesionForcesSum() {
  if (mvCohesionForcesSum.sqrLength() < FloatPrec) calculateCohesionForces();
  return mvCohesionForcesSum;
}

struct Writer: public unary_function<FeFace *, void> {
  Writer(ostream &os):outOs(os) {};
  ostream &outOs;
  void operator () (FeFace *pf) { pf->writeToFile(outOs); };
};

void FeBody::writeToFile(ostream &os) {
  os << "  B{" << endl;
  
  Writer pw(os);
  for_each(mvpFaces.begin(), mvpFaces.end(), pw);

  os << "  }" << endl;
}

struct Informer: public unary_function<FeFace *, void> {
  Informer(ostream &os):outOs(os) {};
  ostream &outOs;
  void operator () (FeFace *pf) { pf->info(outOs); };
};

void FeBody::info(ostream &os) {
  os << "  Body info {" << endl;
  
  Informer pinf(os);
  for_each(mvpFaces.begin(), mvpFaces.end(), pinf);
  
  os << "    Center: " << center() << " Volume: " << volume()
    //     << " Percolated volume: " << percolatedVolume()
     << endl;

  os << "    Density: " << density() << " Weight: " << weight() << endl;

  //  os << "    Initial velocity: " << initialVelocity() << endl;
  os << "    External forces on this body: " << externalForces() << endl;

  os << "    Equilibrium force direction: " << equilibriumForceDirection() << endl;

  os << "  }" << endl;
}

/*--- End of file ---*/
