
#include <iostream>
#include <iomanip>

#include "../libNewMat/newmat.h"
#include "../libNewMat/newmatio.h"

#include "basic.h"
#include "equations.h"
#include "exception.h"

FeSetOfLinearEquations::FeSetOfLinearEquations() {
  mvNOfEquations=0;
  mvSolved=false;
}
FeSetOfLinearEquations::~FeSetOfLinearEquations() {
  map<string, vector<double>* >::iterator actIt=mvColumns.begin();
  while (actIt != mvColumns.end()) {
    delete (*actIt).second;
    actIt++;
  }  
}

unsigned int FeSetOfLinearEquations::newEquations(size_t num) {
  mvSolved=false;

  mvNOfEquations += num;
  
  map<string, vector<double>* >::iterator actIt=mvColumns.begin();
  while (actIt != mvColumns.end()) {
    (*actIt).second->resize(mvNOfEquations);
    actIt++;
  }
  mvEqualsTo.resize(mvNOfEquations);

  return mvNOfEquations - num;
}

double& FeSetOfLinearEquations::operator () (size_t row) {
  if (row > mvNOfEquations - 1) throw FeError("double & FeSetOfLinearEquations::operator () (size_t row): row out of range");

  mvSolved=false;

  return mvEqualsTo[row];
}

double& FeSetOfLinearEquations::operator () (string col, size_t row) {
  if (row > mvNOfEquations - 1) throw FeError("double & FeSetOfLinearEquations::operator () (string col, size_t row): row out of range");

  mvSolved=false;

  vector<double> *actCol=mvColumns[col];
  if (!actCol) {
    actCol=new vector<double>(mvNOfEquations);
    mvColumns[col]=actCol;
  }

  return (*actCol)[row];
}

void FeSetOfLinearEquations::solve() {
  if (!mvNOfEquations) throw FeError("No equations to solve");
  if (mvNOfEquations != mvColumns.size()) {
    cout << "Equations: " << mvNOfEquations << " Variables: " << mvColumns.size() << endl;
    throw FeError("Cannot solve this set of linear equations, number of equations ( != number of variables");
  }

  try {
    Matrix D(mvNOfEquations, mvNOfEquations);
    map<string, vector<double>* >::iterator actIt=mvColumns.begin();
    unsigned int row, col=0;
    while (actIt != mvColumns.end()) {
      for (row=0; row < mvNOfEquations; row ++)
      	D(row+1, col+1)=(*(*actIt).second)[row];           

      actIt++; col++;
    }

    DEBUG(6, cout << endl << "D: " << endl << D << endl);

    ColumnVector E(mvNOfEquations);
    for (row=0; row < mvNOfEquations; row ++)
      E(row+1)=mvEqualsTo[row];

    DEBUG(6, cout << "R: " << endl << E << endl);

    LinearEquationSolver S=D;

    DEBUG(6, cout << "Det: " << S.Determinant() << endl);

    ColumnVector R=S.i() * E;
    
    DEBUG(6, cout << endl << "Result: " << endl << R << endl);
    
    actIt=mvColumns.begin();
    col=0;
    while (actIt != mvColumns.end()) {
      mvResults[(*actIt).first]=R(col + 1);
      actIt++; col++;
    }
  }
  catch (Exception) {
    DEBUG(9, cout << Exception::what() << endl);
    throw FeError("Solve of set of linear equations failed");
  }

  mvSolved=true;
}

double FeSetOfLinearEquations::operator [] (string name) {
  if (!mvSolved) throw FeError("You must solve set of equation before accesing results");
  
  map<string, double>::const_iterator res=mvResults.find(name);
  if (res == mvResults.end()) throw FeError("No such result exists");

  return (*res).second;
}


ostream & operator << (ostream &os, FeSetOfLinearEquations &eq) {
  os << "Set of linear equations results:" << endl;
  
  map<string, double>::iterator res=eq.mvResults.begin();
  while (res != eq.mvResults.end()) {
    cout << (*res).first << ": " << (*res).second << endl;

    res++;
  }

  return os;
}

/*--- Enf of file ---*/
