
#include <math.h>

#include <functional>
#include <algorithm>

#include "vertex.h"
#include "face.h"
#include "exception.h"
#include "edge.h"
#include "wing.h"
#include "body.h"
#include "model.h"

#ifndef _THREAD_SAFE
FeVertex *FeFace::mvpLastVertex;
#endif

FeFace::FeFace(): mvNormal(0.0, 0.0, 0.0), mvCenter(0.0, 0.0, 0.0) {
  mvpLastVertex=NULL;

  mvFixed=false;
  mvFrictionAngle=0.0;
  mvCohesion=0.0;

  mvpBody=NULL;
  mvpOppositeFace=NULL;

  mvCenterValid=false;

  mvArea=0.0;
  mvPercolatedArea=-1.0;
}

FeFace::~FeFace() {
  DEBUG(10, cout << "\t\tDestroying face " << id() << endl);
}

string FeFace::eqIdentifier() {
  FeFace *of=oppositeFace();
  if (of) {
    if (of->id() < id()) return of->idS();
    else return idS();
  } else return idS();
}

void FeFace::fixed(bool value) {
  mvFixed=value;
}

int FeFace::number(int newNumber) {
  if (newNumber) mvNumber=newNumber;

  return mvNumber;
}

bool FeFace::fixed() {
  return mvFixed;
}

void FeFace::addVertex(FeVertex *pv) {
  if (mvpLastVertex) {
    FeWing *nw=mvpLastVertex->getWingTo(pv);
    nw->face(this);
    mvpWings.push_back(nw);
  }
  mvpLastVertex=pv;
}

void FeFace::finish() {
  if (mvpWings.empty()) throw FeError("Empty face.");
  
  if (mvpWings.begin() != mvpWings.end()) throw FeError("Face is not closed.");
}

FeVertex &FeFace::firstVertex() {
  mvpActWing=mvpWings.begin();
  return (*mvpActWing)->v0();
}

FeVertex &FeFace::nextVertex() {
  mvpActWing++;
  if (mvpActWing != mvpWings.end()) {
    return (*mvpActWing)->v0();
  } else {
    throw FeListEnd("FeFace::nextVertex()");
  }
}

void FeFace::invalidatePrecomputedValues() {
  mvNormal=FeZeroVector;

  mvCenterValid=false;

  mvArea=0.0;
  mvPercolatedArea=-1.0;
  
  mvpBody->invalidatePrecomputedValues();
}

void FeFace::checkTopology() {
  DEBUG(9, cout << "\tChecking face " << id() << " topology" << endl);
  
  if (mvpWings.size() < 3) isOK(false);

  DEBUG(9, cout << "\tTopology OK" << endl);
}

void FeFace::setBody(FeBody *pb) {
  if (mvpBody) throw FeError("This face already belongs to some body.");
  
  mvpBody=pb;
}

FeBody *FeFace::body() {
  return mvpBody;
}

void FeFace::setOppositeFace(FeFace *pf) {
  if (mvpOppositeFace) throw FeError("This face already has opposite face");

  mvpOppositeFace=pf;
}

FeFace *FeFace::oppositeFace() {
  return mvpOppositeFace;
}

void FeFace::frictionAngle(double angle) {
  mvFrictionAngle=angle;
}

double FeFace::frictionAngle() {
  return mvFrictionAngle;
}

double FeFace::tanFrictionAngle() {
  return tan(PI*frictionAngle()/180.0);
}

double FeFace::sinFrictionAngle() {
  return sin(PI*frictionAngle()/180.0);
}

double FeFace::cosFrictionAngle() {
  return cos(PI*frictionAngle()/180.0);
}

void FeFace::cohesion(double cohesion) {
  mvCohesion=cohesion;
}

double FeFace::cohesion() {
  return mvCohesion;
}

FeVector FeFace::normal() {
  if (mvNormal.sqrLength() < 1e-5) {
    FeVector pu=mvpWings[0]->v0() - mvpWings[1]->v0();
    FeVector pv=mvpWings[0]->v0() - mvpWings[2]->v0();
    mvNormal=pu*pv;
    mvNormal.normalize();
  }

  return mvNormal;
}

FeVector FeFace::center() {
  if (!mvCenterValid) {
    FeVector pc(0.0, 0.0, 0.0);
    FeVertex pv=firstVertex();
    double num=0.0;
    try {
      while (1) {
	pc += pv;
	num++;
	pv=nextVertex();
      }
    } catch (FeListEnd &) {}

    mvCenter=pc/num;
  }

  return mvCenter;
}

double triangleArea(FeVector &a, FeVector &b, FeVector &c) {
  return 0.5*sqrt(Sqr(Det3(a[1], a[2], 1.0,  b[1], b[2], 1.0,  c.y(), c.z(), 1.0)) + 
		  Sqr(Det3(a[2], a[0], 1.0,  b[2], b[0], 1.0,  c.z(), c.x(), 1.0)) + 
		  Sqr(Det3(a[0], a[1], 1.0,  b[0], b[1], 1.0,  c.x(), c.y(), 1.0)));
}

double FeFace::area() {
  if (mvArea == 0.0) {
    FeVector pc=center();
    
    FeVertex v0=firstVertex();
    FeVertex v1=nextVertex();
    try {
      while (1) {
	//	mvArea += 0.5*sqrt(Sqr(Det3(v0[1], v0[2], 1.0,  v1[1], v1[2], 1.0,  pc.y(), pc.z(), 1.0)) + 
	//			   Sqr(Det3(v0[2], v0[0], 1.0,  v1[2], v1[0], 1.0,  pc.z(), pc.x(), 1.0)) + 
	//			   Sqr(Det3(v0[0], v0[1], 1.0,  v1[0], v1[1], 1.0,  pc.x(), pc.y(), 1.0)));
	mvArea += triangleArea(v0, v1, pc);
	
	v0=v1;
	v1=nextVertex();
      }
    } catch (FeListEnd &) {}
    v1=firstVertex();
    //    mvArea += 0.5*sqrt(Sqr(Det3(v0[1], v0[2], 1.0,  v1[1], v1[2], 1.0,  pc.y(), pc.z(), 1.0)) + 
    //		       Sqr(Det3(v0[2], v0[0], 1.0,  v1[2], v1[0], 1.0,  pc.z(), pc.x(), 1.0)) + 
    //		       Sqr(Det3(v0[0], v0[1], 1.0,  v1[0], v1[1], 1.0,  pc.x(), pc.y(), 1.0)));
    mvArea += triangleArea(v0, v1, pc);
  }

  return mvArea;
}

FeVector intersectionOfLineWithLevel(FeVector &a, FeVector &b, double level) {
  FeVector dir=b - a;
  double pk=(level - a[2])/dir[2];
  return a + pk*dir;
}

double trianglePercolatedArea(FeVector &a, FeVector &b, FeVector &c, double waterLevel) {
  if (a[2] < waterLevel & b[2] < waterLevel & c[2] < waterLevel) return triangleArea(a, b, c);
  if (a[2] >= waterLevel & b[2] >= waterLevel & c[2] >= waterLevel) return 0.0;
  
  FeVector pa(a), pb(b), pc(c), tmp;
  bool flag;
  do {
    flag=false;
    if (pa[2] > pb[2]) { tmp=pa; pa=pb; pb=tmp; flag=true; }
    if (pb[2] > pc[2]) { tmp=pb; pb=pc; pc=tmp; flag=true; }
  } while (flag);

  if (pa[2] < waterLevel & pb[2] < waterLevel) {
    FeVector int1=intersectionOfLineWithLevel(pa, pc, waterLevel);
    FeVector int2=intersectionOfLineWithLevel(pb, pc, waterLevel);

    return triangleArea(pa, pb, int2) + triangleArea(pa, int1, int2);
  } else {
    FeVector int1=intersectionOfLineWithLevel(pa, pb, waterLevel);
    FeVector int2=intersectionOfLineWithLevel(pa, pc, waterLevel);

    return triangleArea(pa, int1, int2);
  }
}

double FeFace::percolatedArea() {
  if (mvPercolatedArea < 0.0) {
    mvPercolatedArea=0.0;

    double wl=mvpBody->model()->waterLevel();
    
    FeVector pc=center();
    
    FeVertex v0=firstVertex();
    FeVertex v1=nextVertex();
    try {
      while (1) {
	mvPercolatedArea += trianglePercolatedArea(v0, v1, pc, wl);
	
	v0=v1;
	v1=nextVertex();
      }
    } catch (FeListEnd &) {}
    v1=firstVertex();
    mvPercolatedArea += trianglePercolatedArea(v0, v1, pc, wl);
  }

  return mvPercolatedArea;
}

FeVector FeFace::cohesionForce() {  
  double value=area() * cohesion();

  FeFace *of=oppositeFace();
  if (of) {
    FeVector velocity=body()->velocity() - of->body()->velocity();
    velocity.normalize();
    return -value*velocity;
  } else if (fixed()) {
    FeVector velocity=body()->velocity();
    velocity.normalize();
    return -value*velocity;
  } else return FeZeroVector;
}

FeVector FeFace::frictionForceDir() {
  FeFace *of=oppositeFace();  
  if (of) {    
    assert(body() != of->body());

    FeVector nv=normal();
    FeVector vv=body()->velocity();
    FeVector ov=of->body()->velocity();
    FeVector dv=vv - ov;
    dv.normalize();
    
    if(dv.sqrLength() < FloatPrec) throw FeException("Two bodies with zero relative velocity detected");

    FeVector ff=-1.0*nv*cosFrictionAngle() - dv*sinFrictionAngle();
    //    FeVector ff=-1.0*nv - tanFrictionAngle()*dv;
    //    ff.normalize();

    return ff;
  } else if (fixed()) {

    FeVector nv=normal();
    FeVector vv=body()->velocity();
    vv.normalize();

    FeVector ff=-1.0*nv*cosFrictionAngle() - vv*sinFrictionAngle();
    //    FeVector ff=-1.0*nv - tanFrictionAngle()*vv;
    //    ff.normalize();

    return ff;

  } else return FeZeroVector;
}

double FeFace::vertexDist(FeVector &pv) {
  FeVector pn=normal();
  FeVector fv=firstVertex();
  double pk=-(pn.x()*fv[0] + pn.y()*fv[1] + pn.z()*fv[2]);
  return pn.x()*pv.x() + pn.y()*pv.y() + pn.z()*pv.z() + pk;
}

void FeFace::writeToFile(ostream &os) {
  os << "    F{"; 

  FeVertex pv=firstVertex();
  try {
    while (1) {
      os << pv;
      pv=nextVertex();
    }
  } catch (FeListEnd &) {}
  os << firstVertex();
 
  os << "}" << endl;
}

void FeFace::info(ostream &os) {
  os << "    Face { ";
  FeVertex pv=firstVertex();
  try {
    while (1) {
      os << pv << " ";
      pv=nextVertex();
    }
  } catch (FeListEnd &) {}
  os << endl << "           Center: " << center() << " Normal: " << normal() << endl;
  os << "           Area: " << area()
    //     << " Percolated area: " << percolatedArea()
     << endl;

  os <<         "           Fixed: " << fixed() << " Cohesion: " << cohesion() << " Friction angle: " << frictionAngle() << " }" << endl;
}

/*--- End of file ---*/
