
#include <stdio.h>
#include <math.h>

#include <functional>
#include <algorithm>

#include "model.h"
#include "exception.h"
#include "equations.h"

FeModel::FeModel() {
  mvpActBody=NULL;
  mvpActFace=NULL;

  DEBUG(10, cout << "New model " << id() << endl);
}

FeModel::~FeModel() {
  DEBUG(10, cout << "Destroying model " << id() << endl);

  DEBUG(10, cout << "\tDestroying bodies" << endl);
  KillFromVector<FeBody> pk1(mvpBodies);
  for_each(mvpBodies.begin(), mvpBodies.end(), pk1);
  mvpBodies.clear();

  DEBUG(10, cout << "\tDestroying faces" << endl);
  KillFromVector<FeFace> pk2(mvpFaces);
  for_each(mvpFaces.begin(), mvpFaces.end(), pk2);
  mvpFaces.clear();

  DEBUG(10, cout << "\tDestroying edges" << endl);
  KillFromVector<FeEdge> pk3(mvpEdges);
  for_each(mvpEdges.begin(), mvpEdges.end(), pk3);
  mvpEdges.clear();

  DEBUG(10, cout << "\tDestroying vertices" << endl);
  KillFromVector<FeVertex> pk4(mvpVertices);
  for_each(mvpVertices.begin(), mvpVertices.end(), pk4);
  mvpVertices.clear();
}

int FeModel::addVertex(FeVector &pv) {
  FeVertex *nv=new FeVertex(pv);
  nv->mvpModel=this;
  mvpVertices.push_back(nv);
  return mvpVertices.size() - 1;
}

FeVertex *FeModel::getVertex(int index) {
  return mvpVertices[index];
}

void FeModel::changeVertex(int index, FeVector &newVertex) {
  (*mvpVertices[index])[0]=newVertex[0];
  (*mvpVertices[index])[1]=newVertex[1];
  (*mvpVertices[index])[2]=newVertex[2];
  mvpVertices[index]->invalidatePrecomputedValues();
}

void FeModel::changeVertex(int index, double x, double y, double z) {
  (*mvpVertices[index])[0]=x;
  (*mvpVertices[index])[1]=y;
  (*mvpVertices[index])[2]=z;
  mvpVertices[index]->invalidatePrecomputedValues();
}

void FeModel::setPertubatedVertices(const int number, const int *indexes) {
  for (int i=0; i < number; i++, indexes++)
    mvPertubatedVertices.push_back((*indexes) - 1);
}

FeEdge *FeModel::createEdge(FeVertex *v0, FeVertex *v1) {
  FeEdge *ne=new FeEdge(v0, v1);
  mvpEdges.push_back(ne);
  return ne;
}

FeBody *FeModel::startBody() {
  mvpActBody=new FeBody();
  mvpActBody->mvpModel=this;

  DEBUG(10, cout << "Starting body " << mvpActBody->id() << endl);

  return mvpActBody;
}

void FeModel::endBody() {
  if (!mvpActBody) throw FeError("Call of FeModel::endBody() without FeModell::startBody()");
  
  mvpActBody->checkTopology();  
  mvpBodies.push_back(mvpActBody);

  DEBUG(10, cout << "Finishing body " << mvpActBody->id() << endl);

  mvpActBody=NULL;
}

FeFace *FeModel::faceStart() {
  mvpActFace=new FeFace();

  DEBUG(10, cout << "Starting face " << mvpActFace->id() << endl);

  return mvpActFace;
}

FeVertex *FeModel::faceAddVertex(int index) {
  mvpActFace->addVertex(mvpVertices[index]);

  return mvpVertices[index];
}

void FeModel::faceEnd() {
  if (!mvpActFace) throw FeError("Call of FeModel::faceEnd() without FeModell::faceStart()");
  
  mvpFaces.push_back(mvpActFace);
  
  mvpActBody->addFace(mvpActFace);

  DEBUG(10, cout << "Finishing face " << mvpActFace->id() << endl);

  mvpActFace=NULL;  
}

FeFace *FeModel::lookupBodyBodyFace(int bodyNum1, int bodyNum2) {
  vector<FeFace *>::iterator actF=mvpFaces.begin();
  while (actF != mvpFaces.end()) {
    if ((*actF)->body()->number() == bodyNum1)
      if ((*actF)->oppositeFace())
	if ((*actF)->oppositeFace()->body()->number() == bodyNum2)
	  return *actF;

    actF++;
  }

  throw FeError("Iterface between this two bodies does not exists");
}

FeFace *FeModel::lookupFace(int bodyNum, int faceNum) {
  vector<FeFace *>::iterator actF=mvpFaces.begin();
  while (actF != mvpFaces.end()) {
    if ((*actF)->body()->number() == bodyNum && (*actF)->number() == faceNum) {
      if ((*actF)->fixed()) return *actF;
      else throw FeError("This face is not with connetion with ground");
    }

    actF++;
  }

  throw FeError("Lookup face failed");
}

double FeModel::waterLevel(double newWaterLevel) {
  if (newWaterLevel != FE_INFINITY) mvWaterLevel=newWaterLevel;
  return mvWaterLevel;
}

/*
FeBody *FeModel::addTetrahedron(FeVertex *v0, FeVertex *v1, FeVertex *v2, FeVertex *v3, bool *stateList) {
  FeVector pu=*v1 - *v0;
  FeVector pv=*v2 - *v0;
  FeVector pw=*v3 - *v0;
  FeVector norm1=pu*pv;
  double pk=norm1^pw;
  if (pk > 0) throw FeError("Bad tetrahedron");

  FeBody *pb=startBody();

  FeFace *pf=faceStart();
  faceAddVertex(v0);
  faceAddVertex(v1);
  faceAddVertex(v2);
  faceAddVertex(v0);
  faceEnd();
  if (stateList) pf->fixed(stateList[0]);

  pf=faceStart();
  faceAddVertex(v0);
  faceAddVertex(v3);
  faceAddVertex(v1);
  faceAddVertex(v0);
  faceEnd();
  if (stateList) pf->fixed(stateList[1]);

  pf=faceStart();
  faceAddVertex(v0);
  faceAddVertex(v2);
  faceAddVertex(v3);
  faceAddVertex(v0);
  faceEnd();
  if (stateList) pf->fixed(stateList[2]);

  pf=faceStart();
  faceAddVertex(v1);
  faceAddVertex(v3);
  faceAddVertex(v2);
  faceAddVertex(v1);
  faceEnd();
  if (stateList) pf->fixed(stateList[3]);

  endBody();

  return pb;
}

FeBody *FeModel::addTetrahedron(double points[4][3], bool *stateList) {
  FeVertex *v0=getVertex(points[0]);
  FeVertex *v1=getVertex(points[1]);
  FeVertex *v2=getVertex(points[2]);
  FeVertex *v3=getVertex(points[3]);

  return addTetrahedron(v0, v1, v2, v3, stateList);
}

FeBody *FeModel::addTetrahedron(const double p1[3], const double p2[3], const double p3[3], const double p4[3], bool *stateList) {
  double pnts[4][3]={
    { p1[0], p1[1], p1[2] },
    { p2[0], p2[1], p2[2] },
    { p3[0], p3[1], p3[2] },
    { p4[0], p4[1], p4[2] }
  };
  return addTetrahedron(pnts, stateList);
}

FeBody *FeModel::addTetrahedron(const FeVector &p1, const FeVector &p2, const FeVector &p3, const FeVector &p4, bool *stateList) {
  FeVertex *v0=getVertex(p1);
  FeVertex *v1=getVertex(p2);
  FeVertex *v2=getVertex(p3);
  FeVertex *v3=getVertex(p4);

  return addTetrahedron(v0, v1, v2, v3, stateList);
}

FeBody *FeModel::addPrism(unsigned short pNum, double pPoints[MAX_PRISM_POINTS][3], double pHgt, bool *stateList) {
  FeVector pa1(pPoints[0][0], pPoints[0][1], pPoints[0][2]);
  FeVector pb1(pPoints[1][0], pPoints[1][1], pPoints[1][2]);
  FeVector pc1(pPoints[2][0], pPoints[2][1], pPoints[2][2]);
  FeVector pu=pb1 - pa1;
  FeVector pv=pc1 - pa1;
  FeVector pn=pu*pv;
  pn.normalize();

  FeBody *pb=startBody();

  faceStart();
  for (unsigned short i=0; i < pNum; i++)
    faceAddVertex(pPoints[i]);
  faceAddVertex(pPoints[0]);
  faceEnd();

  faceStart();
  for (unsigned short i=0; i < pNum; i++) {
    FeVector av(pPoints[pNum - 1 - i]);
    FeVector av2=av - pn*pHgt;
    faceAddVertex(av2);
  }
  FeVector av(pPoints[pNum - 1]);
  FeVector av2=av - pn*pHgt;
  faceAddVertex(av2);
  faceEnd();
  
  for (unsigned short i=0; i < pNum; i++) {
    unsigned int np=(i < pNum - 1)? i+1 : 0;
    FeVector av(pPoints[i]);
    FeVector nv(pPoints[np]);
    FeVector av2=av - pn*pHgt;
    FeVector nv2=nv - pn*pHgt;

    faceStart();
    faceAddVertex(av);
    faceAddVertex(av2);
    faceAddVertex(nv2);
    faceAddVertex(nv);
    faceAddVertex(av);
    faceEnd();
  }

  endBody();

  return pb;
}

FeBody *FeModel::addPyramid(unsigned short pNum, double pPoints[MAX_PRISM_POINTS][3], double pHgt, bool *stateList) {
  FeVector pa1(pPoints[0][0], pPoints[0][1], pPoints[0][2]);
  FeVector pb1(pPoints[1][0], pPoints[1][1], pPoints[1][2]);
  FeVector pc1(pPoints[2][0], pPoints[2][1], pPoints[2][2]);
  FeVector pu=pb1 - pa1;
  FeVector pv=pc1 - pa1;
  FeVector pn=pu*pv;
  pn.normalize();

  FeBody *pb=startBody();

  FeFace *pf=faceStart();
  for (unsigned short i=0; i < pNum; i++)
    faceAddVertex(pPoints[i]);
  faceAddVertex(pPoints[0]);
  faceEnd();

  FeVector cv=pf->center();
  FeVector top=cv - pn*pHgt;

  for (unsigned short i=0; i < pNum; i++) {
    unsigned int np=(i < pNum - 1)? i+1 : 0;
    FeVector av(pPoints[i]);
    FeVector nv(pPoints[np]);

    faceStart();
    faceAddVertex(av);
    faceAddVertex(top);
    faceAddVertex(nv);
    faceAddVertex(av);
    faceEnd();
  }

  endBody();

  return pb;
}
*/

struct CenterChecker: public unary_function<FeFace *, bool> {
  CenterChecker(FeVector &center):mCenter(center) {};
  FeVector &mCenter;
  bool operator () (FeFace *pf) {
    FeVector pc=pf->center();
    FeVector dist=pc - mCenter;
    if (dist.sqrLength() < FE_FLOAT_PREC) return true;
    else return false;
  }
};

void FeModel::detectSharedFaces() {
  vector<FeFace *>::iterator pf=mvpFaces.begin();
  while (pf != mvpFaces.end()) {
    if (!(*pf)->oppositeFace()) {
      FeVector c1=(*pf)->center();
      CenterChecker checker(c1);
      vector<FeFace *>::iterator found=find_if(pf+1, mvpFaces.end(), checker);
      if (found != mvpFaces.end()) {
	(*pf)->setOppositeFace(*found);
	(*found)->setOppositeFace(*pf);
      }
    }

    pf++;
  }
}

void FeModel::solveVelocities() {
  DEBUG(4, cout << "FeModel::solveVelocities()" << endl);

  try {
    FeSetOfLinearEquations eq;     

    DEBUG(6, cout << "Creating set of linear equations" << endl);
   
    // Create set of linear equations
    vector<FeBody *>::iterator actB=mvpBodies.begin();
    while (actB != mvpBodies.end()) {
      FeBody *pb=*actB;
      string pbName=pb->idS();

      vector<FeFace *>::iterator actF=pb->firstFace();
      while (actF != pb->lastFace()) {
	FeFace *pf=*actF;
	FeFace *of=pf->oppositeFace();
	FeVector norm=pf->normal();

	if (pf->fixed() || (of && pf->id() < of->id())) {
	  unsigned int eq1=eq.newEquations(1);
 
	  if (actB == mvpBodies.begin()) {
	    eq(eq1) = norm.z();
	  } else {
	    eq(pbName + "z", eq1) = norm.z();
	    eq(eq1) = 0.0;
	  }
	  eq(pbName + "x", eq1) = norm.x();
	  eq(pbName + "y", eq1) = norm.y();

	  if (of) {
	    string obName=of->body()->idS();

	    eq(obName + "x", eq1) = -norm.x();
	    eq(obName + "y", eq1) = -norm.y();
	    eq(obName + "z", eq1) = -norm.z();
	  }
	}

	actF++;
      }
      
      actB++;
    }

    DEBUG(7, cout << eq << endl);

    // Solve it
    eq.solve();
    
    // And get the results
    DEBUG(3, cout << endl);
    actB=mvpBodies.begin();
    while (actB != mvpBodies.end()) {
      FeBody *pb=*actB;
      string pbName=pb->idS();
      
      if (actB == mvpBodies.begin()) {
	FeVector dv(eq[pbName + "x"], eq[pbName + "y"], -1.0);
	pb->velocity(dv);

	DEBUG(3, cout << "Body " << pb->number() << " velocity: " << pb->velocity() << endl);
      } else {
	FeVector dv(eq[pbName + "x"], eq[pbName + "y"], eq[pbName + "z"]);
	pb->velocity(dv);

	DEBUG(3, cout << "Body " << pb->number() << " velocity: " << pb->velocity() << endl);
      }

      actB++;
    }  
    DEBUG(3, cout << endl);      
  }
  catch (FeException &exc) {
    DEBUG(3, cout << endl << "Solve of velocity field failed" << endl << exc << endl << endl);
    throw FeSolveFailedException();
  }
}

FeVector& FeModel::calculateCohesionForces() {
  mvCohesionForcesSum=FeVector(0, 0, 0);

  vector<FeBody *>::iterator actB=mvpBodies.begin();
  while (actB != mvpBodies.end()) {
    FeVector actCF=(*actB)->cohesionForcesSum();

    mvCohesionForcesSum += actCF;

    actB++;
  }

  return mvCohesionForcesSum;
}

FeVector& FeModel::calculateGravitationalForces() {
  mvGravitationalForcesSum[0]=0.0;
  mvGravitationalForcesSum[1]=0.0;
  mvGravitationalForcesSum[2]=0.0;

  vector<FeBody *>::iterator actB=mvpBodies.begin();
  while (actB != mvpBodies.end()) {
    FeVector gf(0, -(*actB)->weight(), 0);
    mvGravitationalForcesSum += gf;

    actB++;
  }
  
  return mvGravitationalForcesSum;
}

double FeModel::findEquilibrium() {
  DEBUG(4, cout << "Calculation of equlibrium startes" << endl);
  
  // First we need calculated velocity field
  solveVelocities();

  // Now we can build set of equations for equlibrium of all bodies
  try {
    FeSetOfLinearEquations eq;
    string equilibriumForceIdentifier;

    DEBUG(6, cout << "Building set of equations for equilibrium" << endl);

    vector<FeBody *>::iterator actB=mvpBodies.begin();
    while (actB != mvpBodies.end()) {
      unsigned int eq1=eq.newEquations(3);       // We have three equations for body equilibrium in 3D

      // Right sides of equations
      FeVector gf(0.0, 0.0, -(*actB)->weight()); // Gravitation force afecting this body
      FeVector cf=(*actB)->cohesionForcesSum();  // Sum of cohesion forces at this body
      FeVector xf=(*actB)->externalForces();     // External forces afecting this body
      FeVector rs=gf + cf + xf;
      eq(eq1) = -rs[0];
      eq(eq1 + 1) = -rs[1];
      eq(eq1 + 2) = -rs[2];
      
      // Eqilibrium force on this body
      FeVector ef=(*actB)->equilibriumForceDirection(); // Direction of equlibrium force on this body
      if (ef.sqrLength() > FloatPrec) {
	equilibriumForceIdentifier=(*actB)->idS();

	eq(equilibriumForceIdentifier, eq1) = ef[0];
	eq(equilibriumForceIdentifier, eq1 + 1) = ef[1];
	eq(equilibriumForceIdentifier, eq1 + 2) = ef[2];
      }
      
      // Friction forces on this body
      vector<FeFace *>::iterator actF=(*actB)->firstFace();
      while (actF != (*actB)->lastFace()) {
	FeVector ffdir=(*actF)->frictionForceDir();

	if (ffdir.sqrLength() > FloatPrec ) {
	  string fi=(*actF)->eqIdentifier(); // Identifier of this face for equations (for concident faces is the same identifier!)

	  eq(fi, eq1) = ffdir[0];
	  eq(fi, eq1 + 1) = ffdir[1];
	  eq(fi, eq1 + 2) = ffdir[2];
	}
	
	actF++;
      }

      actB++;
    }    

    DEBUG(6, cout << "Solving it" << endl);
    
    // Solve it
    eq.solve();

    DEBUG(6, cout << "Solving done" << endl);

    // Print the results
    DEBUG(3, cout << endl);    
    DEBUG(3, 
	  vector<FeFace *>::iterator actF=mvpFaces.begin();
	  while (actF != mvpFaces.end()) {
	    FeFace *of=(*actF)->oppositeFace();
	    if ((*actF)->oppositeFace()) {
	      string idS=(*actF)->eqIdentifier();
	      cout << "Cohesion force on interface " << (*actF)->body()->number() << " <-> " << of->body()->number()
		   << ": " << (*actF)->cohesionForce() << endl;
	    } else if ((*actF)->fixed()) {
	      string idS=(*actF)->eqIdentifier();
	      cout << "Cohesion force on interface " << (*actF)->body()->number() << " <-> " << "ground"
		   << ": " << (*actF)->cohesionForce() << endl;
	    }
	    
	    actF++;
	  }
	  );
    DEBUG(3, cout << endl);
    DEBUG(3, 
	  vector<FeFace *>::iterator actF=mvpFaces.begin();
	  while (actF != mvpFaces.end()) {
	    FeFace *of=(*actF)->oppositeFace();
	    if (of) {
	      string idS=(*actF)->eqIdentifier();
	      cout << "Friction force on interface " << (*actF)->body()->number() << " <-> " << of->body()->number()
		   << " ... Value: " << eq[idS] << " Direction: " << (*actF)->frictionForceDir() << endl;
	    } else if ((*actF)->fixed()) {
	      string idS=(*actF)->eqIdentifier();
	      cout << "Friction force on interface " << (*actF)->body()->number() << " <-> " << "ground"
		   << " ... Value: " << eq[idS] << " Direction: " << (*actF)->frictionForceDir() << endl;
	    }
	    
	    actF++;
	  }
	  );

    DEBUG(1, cout << endl << "Equilibrium force value: " << eq[equilibriumForceIdentifier] << endl << endl);

    return eq[equilibriumForceIdentifier];
  }
  catch (FeException &exc) {
    DEBUG(1, cout << endl << "Solve of equilibrium failed" << endl << exc << endl << endl);
    throw FeSolveFailedException();
  }  
}

bool FeModel::pertubVertices(int iterationNumber, bool lastIterationOK) {
  switch (iterationNumber) {
    case 0:
      changeVertex(0, 0.0,  0.0, 0.0);
      changeVertex(1, 7.0,  0.0, 7.0);
      changeVertex(2, 0.0,  0.0, 7.0);
      changeVertex(3, 0.0, 15.0, 7.0);
      break;
    case 1:
      changeVertex(0,  0.0,  0.0,  0.0);
      changeVertex(1, 10.0,  0.0, 10.0);
      changeVertex(2,  0.0,  0.0, 10.0);
      changeVertex(3,  0.0, 15.0, 17.0);
      break;
    case 2:
      changeVertex(0,  0.0, -3.0,  0.0);
      changeVertex(1,  7.0,  0.0, 10.0);
      changeVertex(2, -7.0,  0.0, 10.0);
      changeVertex(3,  0.0, 16.0, 10.0);
      break;
    case 3:
      changeVertex(0,  0.0,  0.0,  0.0);
      changeVertex(1,  6.0,  0.0, 10.0);
      changeVertex(2, -6.0,  0.0, 10.0);
      changeVertex(3,  0.0, 18.0, 25.0);
      break;
  }

  return true;
}

double FeModel::findMinimalEqulibriumForce(int maxIterations) {
  mvMinimalEFValue=1e20;

  for (int iteration=0; iteration < maxIterations; iteration++) {
    DEBUG(0.5, cout << "Iteration " << iteration << ": ");

    try {
      double actEFValue=findEquilibrium();
      DEBUG(0.5, cout << actEFValue << endl);

      if (CHECK_IF_NEW_EQ_FORCE_IS_BETTER(actEFValue, mvMinimalEFValue)) {
	mvMinimalEFValue=actEFValue;
	mvMinimalEFVCoords.clear();
	for (vector<FeVertex *>::iterator actV=mvpVertices.begin(); actV < mvpVertices.end(); actV++)
	  mvMinimalEFVCoords.push_back(*actV);
	
	mvMinimalEFNumIt=iteration;
      }
      
      if (!pertubVertices(iteration, true)) break;
    }
    catch (FeSolveFailedException) {
      DEBUG(0.5, cout << "Solve failed" << endl);
      if (!pertubVertices(iteration, false)) break;
    }
  }
  
  for (unsigned int i=0; i < mvpVertices.size(); i++)
    changeVertex(i, mvMinimalEFVCoords[i]);

  DEBUG(0.5, cout << "Minimal equlibrium force value: " << mvMinimalEFValue << " (iteration number " << mvMinimalEFNumIt << ")" << endl);  

  return mvMinimalEFValue;
}

void performCheck(FeBasic *pb) {
  pb->checkTopology();
}

void FeModel::checkTopology() {
  DEBUG(9, cout << "Checking model " << id() << " topology" << endl);

  for_each(mvpEdges.begin(), mvpEdges.end(), performCheck);
  for_each(mvpFaces.begin(), mvpFaces.end(), performCheck);
  for_each(mvpBodies.begin(), mvpBodies.end(), performCheck);

  DEBUG(9, cout << "\t" << mvpVertices.size() << " vertices, " << mvpEdges.size() << " edges, " << mvpFaces.size() << " faces, " << mvpBodies.size() << " bodies" << endl);

  detectSharedFaces();
}

struct Writer: public unary_function<FeBody *, void> {
  Writer(ostream &os):outOs(os) {};
  ostream &outOs;
  void operator () (FeBody *pb) { pb->writeToFile(outOs); };
};

void FeModel::writeToFile(ostream &os) {
  os << "M{" << endl;
  
  Writer pw(os);
  for_each(mvpBodies.begin(), mvpBodies.end(), pw);

  os << "}" << endl << endl;
}

struct Informer: public unary_function<FeBody *, void> {
  Informer(ostream &os):outOs(os) {};
  ostream &outOs;
  void operator () (FeBody *pb) { pb->info(outOs); };
};

void FeModel::info(ostream &os) {
  os << endl << "Model info { " << endl;

  //  os << "  Water level: " << waterLevel() << endl;

  os << "  Model vertices {" << endl;
  for (unsigned int i=0; i < mvpVertices.size(); i++)
    os << "    " << i+1 << ". " << mvpVertices[i] << endl;
  os << "  }" << endl;

  Informer pinf(os);
  for_each(mvpBodies.begin(), mvpBodies.end(), pinf);  

  //  os << "  Pertubated vertices: ";
  //  for (vector<int>::iterator pv=mvPertubatedVertices.begin(); pv != mvPertubatedVertices.end(); pv++)
  //    os << *pv + 1 << " ";
  //  os << endl;

  os << "}" << endl << endl;
}

/*--- End of file ---*/
