
#include <stdio.h>
#include <string.h>

#include "exception.h"

#include "parser.h"

char voidChars[]=" \n\t;";

FeParser::FeParser(FeModel *model, FILE *inputFile) {
  mvInputFile=inputFile;

  mvpActModel=model;
  mvpActBody=NULL;
  mvNumOfBodies=0;

  mvActLine[0]=0;
  
  getModel();
}

bool FeParser::readLine(int mode) {
  mvActLine[0]=0;

  switch (mode) {
    case 0:
      while (strchr(voidChars, mvActLine[0]) || !mvActLine[0]) {
	if (!fgets(mvActLine, MAX_LINE_LEN, mvInputFile)) {
	  mvActLine[0]=0;
	  break;
	}
      }
      break;
    case 1:
      while (strchr(";\n\t", mvActLine[0]) || !mvActLine[0]) {
	if (!fgets(mvActLine, MAX_LINE_LEN, mvInputFile)) {
	  
	  mvActLine[0]=0;
	  break;
	}
      }
      break;
  }
      
  mvpActChar=mvActLine;
  mvActLine[strlen(mvActLine) - 1]=0;
    
  DEBUG(12, cout << "Processing line '" << mvActLine << "'" << endl);

  char *ps=mvActLine;
  while (strchr(voidChars, *ps))
    if (!(*(ps++))) return false;

  return (mvActLine[0] != 0);
}

int FeParser::getInteger() {
  readLine();

  int retVal;
  if (sscanf(mvpActChar, "%d", &retVal) < 1) throw FeError("Cannot read integer");
  
  return retVal;
}

double FeParser::getFloat() {
  readLine();

  double retVal;
  if (sscanf(mvpActChar, "%lf", &retVal) < 1) throw FeError("Cannot read float");

  return retVal;
}

bool FeParser::getBoolean() {
  readLine();

  bool retVal;
  if (!strncmp(mvpActChar, "yes", 3)) retVal=true;
  else if (!strncmp(mvpActChar, "no", 2)) retVal=false;
  else throw FeError("Cannot read boolean");

  return retVal;
}

const char *FeParser::getIdentifier() {
  readLine();

  char *ps=mvBuffer;
  
  while (isalpha(*mvpActChar))
    *(ps++) = *(mvpActChar++);

  *ps=0;

  if (!mvBuffer[0]) throw FeError("Cannot read identifier");

  return mvBuffer;
}

FeVector FeParser::getVector() {
  readLine();

  double x, y, z;
  if (sscanf(mvpActChar, "%lf %lf %lf", &x, &y, &z) < 3) throw FeError("Cannot read vector");

  return FeVector(x, y, z);
}

int FeParser::getListOfIntegers() {
  readLine();
  
  mvNumberOfIntegersInList=0;
  while (sscanf(mvpActChar, "%d", mvListOfIntegers + mvNumberOfIntegersInList) == 1) {
    mvNumberOfIntegersInList++;
    mvpActChar=strchr(mvpActChar, ' ');
    
    if (!mvpActChar) break;

    mvpActChar++;
  }
 
  if (!mvNumberOfIntegersInList) throw FeError("Cannot read list of integers");

  return mvNumberOfIntegersInList;
}

int *FeParser::copyListOfIntegers() {
  int *newL=(int *)malloc(mvNumberOfIntegersInList*sizeof(int));
  int *fi=mvListOfIntegers, *ti=newL;

  for (int i=0; i < mvNumberOfIntegersInList; i++, fi++, ti++)
    *ti=*fi;

  return newL;
}

void FeParser::getFacesDetails(int num, FeFace **faces) {
  int numi=getListOfIntegers();
  for (int i=0; i < numi; i++) {
    if (mvListOfIntegers[i] < 0 || mvListOfIntegers[i] > num) throw FeError("Bad index of face");
    if (mvListOfIntegers[i])
      faces[mvListOfIntegers[i] - 1]->fixed(true);
  }

  for (int i=0; i < num; i++)
    faces[i]->number(i + 1);
}

void FeParser::getBodyDetails(FeBody *pb) {
  pb->density(getFloat());

  FeVector ef=getVector();
  pb->externalForces(ef);

  FeVector efd=getVector();
  pb->equilibriumForceDirection(efd);
}

void FeParser::getInterfaceProperties() {
  int body1, body2, face;
  while (readLine(1)) {
    char *ps=strdup(mvpActChar), *os=ps;

    if (!(ps=strtok(ps, voidChars))) throw FeError("Bad interface properties definition [1]");
    body1=atoi(ps);
    if (!(ps=strtok(NULL, voidChars))) throw FeError("Bad interface properties definition [3]");
    body2=atoi(ps);

    if (!body1 || !body2) {
      if (!(ps=strtok(NULL, voidChars))) throw FeError("Bad interface properties definition [5]");
      if (!(face=atoi(ps))) throw FeError("Bad interface properties definition [6]");
    }

    if (!(ps=strtok(NULL, voidChars))) throw FeError("Bad interface properties definition [7]");
    double fa=atof(ps);

    if (!(ps=strtok(NULL, voidChars))) throw FeError("Bad interface properties definition [8]");
    double co=atof(ps);
    
    FeFace *pf;
    if (body1 && body2) pf=mvpActModel->lookupBodyBodyFace(body1, body2);
    else pf=mvpActModel->lookupFace(max(body1, body2), face);
    
    pf->frictionAngle(fa);
    pf->cohesion(co);

    if (body1 && body2) {
      pf->oppositeFace()->frictionAngle(fa);
      pf->oppositeFace()->cohesion(co);
    }

    free(os);
  }
}

void FeParser::startBody() {
  mvpActBody=mvpActModel->startBody();
  mvpActBody->number(++mvNumOfBodies);
}

void FeParser::endBody(int numOfFaces, FeFace **faces) {
  mvpActModel->endBody();

  getFacesDetails(numOfFaces, faces);

  getBodyDetails(mvpActBody);
}

FeBody *FeParser::getTetrahedron() {
  DEBUG(11, cout << "Reading tetrahedron" << endl);  

  if (getListOfIntegers() != 4) throw FeError("Tetrahedron should have four vertices");

  int v0=mvListOfIntegers[0] - 1;
  int v1=mvListOfIntegers[1] - 1;
  int v2=mvListOfIntegers[2] - 1;
  int v3=mvListOfIntegers[3] - 1;

  startBody();

  FeFace *pf[4];

  pf[0]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(v0);
  mvpActModel->faceAddVertex(v1);
  mvpActModel->faceAddVertex(v2);
  mvpActModel->faceAddVertex(v0);
  mvpActModel->faceEnd();

  pf[1]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(v0);
  mvpActModel->faceAddVertex(v3);
  mvpActModel->faceAddVertex(v1);
  mvpActModel->faceAddVertex(v0);
  mvpActModel->faceEnd();

  pf[2]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(v0);
  mvpActModel->faceAddVertex(v2);
  mvpActModel->faceAddVertex(v3);
  mvpActModel->faceAddVertex(v0);
  mvpActModel->faceEnd();

  pf[3]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(v1);
  mvpActModel->faceAddVertex(v3);
  mvpActModel->faceAddVertex(v2);
  mvpActModel->faceAddVertex(v1);
  mvpActModel->faceEnd();

  endBody(4, pf);

  return mvpActBody;
}

FeBody *FeParser::getPyramid() {
  DEBUG(11, cout << "Reading pyramid" << endl);  

  int num=getInteger();
  if (getListOfIntegers() != num) throw FeError("Bad number of pyramid base vertices");
  int apex=getInteger();

  startBody();

  FeFace *pf[num + 1];
  
  pf[0]=mvpActModel->faceStart();
  for (int i=0; i < num; i++)
    mvpActModel->faceAddVertex(mvListOfIntegers[i] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[0] - 1);
  mvpActModel->faceEnd();

  for (int i=0; i < num; i++) {
    pf[i + 1]=mvpActModel->faceStart();
    mvpActModel->faceAddVertex(mvListOfIntegers[i] - 1);
    mvpActModel->faceAddVertex(apex - 1);
    mvpActModel->faceAddVertex(mvListOfIntegers[(i < num -1)?(i + 1):0] - 1);
    mvpActModel->faceAddVertex(mvListOfIntegers[i] - 1);
    mvpActModel->faceEnd();
  }
  
  endBody(num + 1, pf);

  return mvpActBody;
}

FeBody *FeParser::getPrism() {
  DEBUG(11, cout << "Reading prism" << endl);  

  int num=getInteger();
  
  if (getListOfIntegers() != num) throw FeError("Bad number of prism1 base vertices");
  int *list0=copyListOfIntegers();
  if (getListOfIntegers() != num) throw FeError("Bad number of prism2 base vertices");

  startBody();

  FeFace *pf[num + 2];
  
  pf[0]=mvpActModel->faceStart();
  for (int i=0; i < num; i++)
    mvpActModel->faceAddVertex(list0[i] - 1);
  mvpActModel->faceAddVertex(list0[0] - 1);
  mvpActModel->faceEnd();
  
  pf[1]=mvpActModel->faceStart();
  for (int i=0; i < num; i++)
    mvpActModel->faceAddVertex(mvListOfIntegers[num - i - 1] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[num - 1] - 1);
  mvpActModel->faceEnd();

  for (int i=0; i < num; i++) {
    pf[i + 2]=mvpActModel->faceStart();
    mvpActModel->faceAddVertex(list0[i] - 1);
    mvpActModel->faceAddVertex(mvListOfIntegers[i] - 1);
    mvpActModel->faceAddVertex(mvListOfIntegers[(i < num -1)?(i + 1):0] - 1);
    mvpActModel->faceAddVertex(list0[(i < num -1)?(i + 1):0] - 1);
    mvpActModel->faceAddVertex(list0[i] - 1);
    mvpActModel->faceEnd();
  }

  endBody(num + 2, pf);

  free(list0);

  return mvpActBody;
}

FeBody *FeParser::getWedge() {
  DEBUG(11, cout << "Reading wedge" << endl);  

  int num=getInteger();
  if (num != 4) throw FeError("Only wedges with four base vertices implemented");
  if (getListOfIntegers() != num) throw FeError("Bad number of wedge base vertices");
  int *list0=copyListOfIntegers();

  if (getListOfIntegers() != 2) throw FeError("Wedge should has two apexes");

  startBody();

  FeFace *pf[num + 1];
  
  pf[0]=mvpActModel->faceStart();
  for (int i=0; i < num; i++)
    mvpActModel->faceAddVertex(list0[i] - 1);
  mvpActModel->faceAddVertex(list0[0] - 1);
  mvpActModel->faceEnd();

  pf[1]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(list0[0] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[0] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[1] - 1);
  mvpActModel->faceAddVertex(list0[1] - 1);
  mvpActModel->faceAddVertex(list0[0] - 1);
  mvpActModel->faceEnd();

  pf[2]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(list0[1] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[1] - 1);
  mvpActModel->faceAddVertex(list0[2] - 1);
  mvpActModel->faceAddVertex(list0[1] - 1);
  mvpActModel->faceEnd();

  pf[3]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(list0[2] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[1] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[0] - 1);
  mvpActModel->faceAddVertex(list0[3] - 1);
  mvpActModel->faceAddVertex(list0[2] - 1);
  mvpActModel->faceEnd();

  pf[4]=mvpActModel->faceStart();
  mvpActModel->faceAddVertex(list0[3] - 1);
  mvpActModel->faceAddVertex(mvListOfIntegers[0] - 1);
  mvpActModel->faceAddVertex(list0[0] - 1);
  mvpActModel->faceAddVertex(list0[3] - 1);
  mvpActModel->faceEnd();

  endBody(num + 1, pf);

  free(list0);

  return mvpActBody;
}

void FeParser::getVertices() {
  int vNum=getInteger();
  for (int i=0; i < vNum; i++) {
    FeVector pv=getVector();
    mvpActModel->addVertex(pv);
  }

  //  getListOfIntegers();
  //  if (!mvNumberOfIntegersInList) throw FeError("Pertubated vertices should be defined (or zero meaning no pertubation)");
  //  if (mvListOfIntegers[0]) mvpActModel->setPertubatedVertices(mvNumberOfIntegersInList, mvListOfIntegers);
}

FeBody *FeParser::getBody() {
  const char *identifier=getIdentifier();

  if (!strcmp(identifier, "tetrahedron")) return getTetrahedron();
  else if (!strcmp(identifier, "pyramid")) return getPyramid();
  else if (!strcmp(identifier, "prism")) return getPrism();
  else if (!strcmp(identifier, "wedge")) return getWedge();

  throw FeError("Bad type of body");
}

FeModel *FeParser::getModel() {
  getVertices();

  //  double wl=getFloat();
  //  if (wl <= -10000) wl=FE_INFINITY;
  //  mvpActModel->waterLevel(wl);

  int nOfBodies=getInteger();
  if (nOfBodies < 1) throw FeError("At least one body should be defined");
  
  for (int i=0; i < nOfBodies; i++) {
    FeBody *actBody=getBody();
    actBody->checkTopology();
  }

  mvpActModel->checkTopology();

  getInterfaceProperties();

  return mvpActModel;
}

/*--- End of file ---*/
