
#include <math.h>
#include <assert.h>

#include "vector.h"
#include "vertex.h"

#ifndef Det2
#define Det2(a, b, c, d) ((a)*(d) - (b)*(c))
#endif

#ifndef Det3
#define Det3(a,b,c, d,e,f, g,h,i) ((a)*(e)*(i) + (d)*(h)*(c) + (g)*(b)*(f) - (g)*(e)*(c) - (d)*(b)*(i) - (a)*(h)*(f))
#endif

FeVector::FeVector() {
  mvA[0]=0.0;
  mvA[1]=0.0;
  mvA[2]=0.0;
}

FeVector::FeVector(double *a) {
  mvA[0]=a[0];
  mvA[1]=a[1];
  mvA[2]=a[2];
}

FeVector::FeVector(double x, double y, double z) {
  mvA[0]=x;
  mvA[1]=y;
  mvA[2]=z;
}

FeVector::FeVector(const FeVector &pv) {
  mvA[0]=pv.mvA[0];
  mvA[1]=pv.mvA[1];
  mvA[2]=pv.mvA[2];
}

FeVector::FeVector(const FeVector *pv) {
  mvA[0]=pv->mvA[0];
  mvA[1]=pv->mvA[1];
  mvA[2]=pv->mvA[2];
}

double& FeVector::operator [] (int a) {
  assert(a >= 0 && a < 4);
  return mvA[a];
}

///////////////////

FeVector &FeVector::operator = (const FeVector &v) {
  if (&v != this) { mvA[0]=v.mvA[0]; mvA[1]=v.mvA[1]; mvA[2]=v.mvA[2]; }
  return *this;
}

FeVector &FeVector::operator += (const FeVector &v) {
  mvA[0]+=v.mvA[0];
  mvA[1]+=v.mvA[1];
  mvA[2]+=v.mvA[2];
  return *this;
}

FeVector &FeVector::operator -= (const FeVector &v) {
  mvA[0]-=v.mvA[0];
  mvA[1]-=v.mvA[1];
  mvA[2]-=v.mvA[2];
  return *this;
}

FeVector FeVector::operator + (const FeVector &v) const {  
  return FeVector(mvA[0] + v.mvA[0], mvA[1] + v.mvA[1], mvA[2] + v.mvA[2]);
}

FeVector FeVector::operator - (const FeVector &v) const {
  return FeVector(mvA[0] - v.mvA[0], mvA[1] - v.mvA[1], mvA[2] - v.mvA[2]);
}

FeVector FeVector::operator * (const double k) const {
  return FeVector(k*mvA[0], k*mvA[1], k*mvA[2]);
}

FeVector FeVector::operator / (const double k) const {
  return FeVector(mvA[0]/k, mvA[1]/k, mvA[2]/k);
}

double FeVector::operator ^ (const FeVector &v) const {
  return mvA[0]*v.mvA[0] + mvA[1]*v.mvA[1] + mvA[2]*v.mvA[2];
}

FeVector FeVector::operator * (const FeVector &v) const {
  return FeVector(Det2(mvA[1], mvA[2], v.mvA[1], v.mvA[2]),
		  -Det2(mvA[0], mvA[2], v.mvA[0], v.mvA[2]),
		  Det2(mvA[0], mvA[1], v.mvA[0], v.mvA[1]));
}

//////////////////////////

double FeVector::sqrLength() {
  return mvA[0]*mvA[0] + mvA[1]*mvA[1] + mvA[2]*mvA[2];
}

double FeVector::length() {
  return sqrt(mvA[0]*mvA[0] + mvA[1]*mvA[1] + mvA[2]*mvA[2]);
}

void FeVector::normalize() {
  double sqrLen=sqrLength();
  if (sqrLen > FloatPrec) {
    if (fabs(1.0 - sqrLen) > FloatPrec) {
      double len=sqrt(sqrLen);
      mvA[0]/=len;
      mvA[1]/=len;
      mvA[2]/=len;
    }
  } else {
    DEBUG(15, cerr << "Warning, normalizing zero vector" << endl);
  }
}

double FeVector::x() {
  return mvA[0];
}

double FeVector::y() {
  return mvA[1];
}

double FeVector::z() {
  return mvA[2];
}

ostream& operator << (ostream &os, const FeVector &pv) {
  return os << "<" << pv.mvA[0] << ", " << pv.mvA[1] << ", " << pv.mvA[2] << ">";
}
 
ostream& operator << (ostream &os, const FeVector *pv) {
  return os << "<" << pv->mvA[0] << ", " << pv->mvA[1] << ", " << pv->mvA[2] << ">";
}

FeVector operator + (FeVertex &v0, FeVertex &v1) {
  return FeVector(v0.mvA[0] + v1.mvA[0], v0.mvA[1] + v1.mvA[1], v0.mvA[2] + v1.mvA[2]);
}

FeVector operator - (FeVertex &v0, FeVertex &v1) {
  return FeVector(v0.mvA[0] - v1.mvA[0], v0.mvA[1] - v1.mvA[1], v0.mvA[2] - v1.mvA[2]);
}

FeVector operator * (double val, FeVector &pv) {  
  return FeVector(val*pv[0], val*pv[1], val*pv[2]);
}


void FeVector::writeToFile(ostream &os) {
  os << this;
}


/*--- End of file ---*/
