
/*
  Oook geometry library
  
  Written in 1998 by Vaclav Cermak (disnel@linux.itam.cas.cz)
  
  Matrix 4x4 code
*/

#include "OGLMatrix4.h"

OGLMatrix4::OGLMatrix4(OGLBool initialize) {
  if (initialize) {
    for (OGLInt i=0;i<16;i++)
      a[i]=0.0;
    a[0]=a[5]=a[10]=a[15]=1.0;
  }
}

OGLMatrix4::OGLMatrix4(const OGLMatrix4 &m) {
  for (OGLInt i=0;i<16;i++)
    a[i]=m.a[i];
}

OGLMatrix4 OGLMatrix4::createIdentity() {
  OGLMatrix4 m;
  m.a[0]=m.a[5]=m.a[10]=m.a[15]=1.0;
  return m;
}

OGLMatrix4 OGLMatrix4::createTranslation(const OGLVector3 &t) {
  OGLMatrix4 m;
  m.a[12]=t[0]; m.a[13]=t[1]; m.a[14]=t[2];
  m.a[0]=m.a[5]=m.a[10]=m.a[15]=1.0;
  return m;
}

OGLMatrix4 OGLMatrix4::createRotationDeg(const OGLVector3 &rd) {
  OGLFloat srx=sin(toRad(rd[0])), crx=cos(toRad(rd[0]));
  OGLFloat sry=sin(toRad(rd[1])), cry=cos(toRad(rd[1]));
  OGLFloat srz=sin(toRad(rd[2])), crz=cos(toRad(rd[2]));

  OGLMatrix4 rx, ry, rz;
  rx.a[5]=crx; rx.a[6]=srx; rx.a[9]=-srx; rx.a[10]=crx; rx.a[0]=rx.a[15]=1.0;
  ry.a[0]=cry; ry.a[2]=-sry; ry.a[8]=sry; ry.a[10]=cry; ry.a[5]=ry.a[15]=1.0;
  rz.a[0]=crz; rz.a[1]=srz; rz.a[4]=-srz; rz.a[5]=crz; rz.a[10]=rz.a[15]=1.0;

  return rx*ry*rz;
}

OGLMatrix4 OGLMatrix4::createRotationRad(const OGLVector3 &rd) {
  OGLFloat srx=sin(rd[0]), crx=cos(rd[0]);
  OGLFloat sry=sin(rd[1]), cry=cos(rd[1]);
  OGLFloat srz=sin(rd[2]), crz=cos(rd[2]);

  OGLMatrix4 rx, ry, rz; 
  rx.a[5]=crx; rx.a[6]=srx; rx.a[9]=-srx; rx.a[10]=crx; rx.a[0]=rx.a[15]=1.0;
  ry.a[0]=cry; ry.a[2]=-sry; ry.a[8]=sry; ry.a[10]=cry; ry.a[5]=ry.a[15]=1.0;
  rz.a[0]=crz; rz.a[1]=srz; rz.a[4]=-srz; rz.a[5]=crz; rz.a[10]=rz.a[15]=1.0;

  return rx*ry*rz;
}

OGLMatrix4 OGLMatrix4::createScale(const OGLVector3 &s) {
  OGLMatrix4 m;
  m.a[0]=s[0]; m.a[5]=s[1]; m.a[10]=s[2]; m.a[15]=1.0;
  return m;
}

OGLMatrix4 OGLMatrix4::createMatrix(const OGLFloat a[16]) {
  OGLMatrix4 m(false);
  for (OGLInt i=0;i<16;i++)
    m.a[i]=a[i];
  return m;
}

OGLMatrix4 OGLMatrix4::operator * (const OGLMatrix4 &m) const {
  OGLMatrix4 nm;
  for (OGLInt i=0;i<4;i++) {
    for (OGLInt j=0;j<4;j++) {
      OGLFloat sum=0.0;
      for (OGLInt k=0;k<4;k++)
	sum+=(*this)(i,k) * m(k,j);
      nm(i,j)=sum;
    }
  }
  return nm;
}

OGLVector4 OGLMatrix4::operator * (const OGLVector4 &v) const {
  return OGLVector4(v[0]*a[0] + v[1]*a[4] + v[2]*a[8] + v[3]*a[12],
		    v[0]*a[1] + v[1]*a[5] + v[2]*a[9] + v[3]*a[13],
		    v[0]*a[2] + v[1]*a[6] + v[2]*a[10] + v[3]*a[14],
		    v[0]*a[3] + v[1]*a[7] + v[2]*a[11] + v[3]*a[15]);
}

OGLVector3 OGLMatrix4::transformVector(const OGLVector3 &v) const {
  return OGLVector3(v[0]*a[0] + v[1]*a[4] + v[2]*a[8],
		    v[0]*a[1] + v[1]*a[5] + v[2]*a[9],
		    v[0]*a[2] + v[1]*a[6] + v[2]*a[10]);
}

OGLVector3 OGLMatrix4::transformVertex(const OGLVector3 &v) const {
  return OGLVector3(v[0]*a[0] + v[1]*a[4] + v[2]*a[8] + a[12],
		    v[0]*a[1] + v[1]*a[5] + v[2]*a[9] + a[13],
		    v[0]*a[2] + v[1]*a[6] + v[2]*a[10] + a[14]);
}

OGLVector4 OGLMatrix4::transformVertex(const OGLVector4 &v) const {
  return OGLVector4(v[0]*a[0] + v[1]*a[4] + v[2]*a[8] + v[3]*a[12],
		    v[0]*a[1] + v[1]*a[5] + v[2]*a[9] + v[3]*a[13],
		    v[0]*a[2] + v[1]*a[6] + v[2]*a[10] + v[3]*a[14],
		    v[0]*a[3] + v[1]*a[7] + v[2]*a[11] + v[3]*a[15]);  
}

OGLFloat OGLMatrix4::operator () (const OGLInt row, const OGLInt col) const {
  return a[4*col + row];
}

OGLFloat &OGLMatrix4::operator () (const OGLInt row, const OGLInt col) {
  return a[4*col + row];
}

ostream &operator << (ostream &o, const OGLMatrix4 &m) {
  o << "[" << m(0,0) << ", " << m(0,1) << ", " << m(0,2) << ", " << m(0,3) << "]\n";
  o << "[" << m(1,0) << ", " << m(1,1) << ", " << m(1,2) << ", " << m(1,3) << "]\n";
  o << "[" << m(2,0) << ", " << m(2,1) << ", " << m(2,2) << ", " << m(2,3) << "]\n";
  o << "[" << m(3,0) << ", " << m(3,1) << ", " << m(3,2) << ", " << m(3,3) << "]\n";
  return o;
}

#ifdef _USE_GL_

void OGLMatrix4::loadGLMatrix() {
#ifdef _USE_GL_DOUBLE_PRECISION_
  glLoadMatrixd(a);
#else
  glLoadMatrixf(a);
#endif  
}

void OGLMatrix4::multGLMatrix() {
#ifdef _USE_GL_DOUBLE_PRECISION_
  glMultMatrixd(a);
#else
  glMultMatrixf(a);
#endif  
}

#endif

/*--- End of file ---*/
