
//#define BUILD_MODEL

#define XK_LATIN1

#include <X11/Xlib.h>
#include <X11/keysymdef.h>

#include <GL/gl.h>
#include <GL/glx.h>

#include <vector>
#include <functional>
#include <algorithm>

#ifdef _OLD_GCC
#include <hash_set>
#else
#include <ext/hash_set>
#endif

#include <FlexLexer.h>

#include "libFailure/libFailure.h"

#include "viewport.h"
#include "glModel.h"
#include "share.h"

struct DrawObject: public unary_function<GLDrawable *, void> {
  DrawObject(GLViewport *pv) { actViewport=pv; };
  GLViewport *actViewport;
  void operator () (GLDrawable *po) { po->glDraw(actViewport); }
};

struct DrawNormal: public unary_function<GLDrawable *, void> {
  DrawNormal(GLViewport *pv) { actViewport=pv; };
  GLViewport *actViewport;
  void operator () (GLDrawable *po) { po->glDrawNormals(actViewport); }
};

class FeViewport: public GLViewport {
private:
  vector<GLDrawable *> mvpObjects;
  unsigned int mvWidth, mvHeight;

public:
  FeViewport(int num, viewportType type, viewportRenderMode mode): GLViewport(num, type, mode) {
  };
  
  void addObject(GLDrawable *po) {
    mvpObjects.push_back(po);
  }
    
  virtual void drawObjects() {
    GLViewport::drawObjects();
    DrawObject pfn(this);
    for_each(mvpObjects.begin(), mvpObjects.end(), pfn);
  };

  virtual void drawNormals() {
    GLViewport::drawNormals();
    DrawNormal pfn(this);
    for_each(mvpObjects.begin(), mvpObjects.end(), pfn);
  };

  virtual void redraw(Display *disp, Window win) {
    if (XPending(disp)) return;

    setRenderingArea(0, 0, mvWidth, mvHeight);
    
    GLViewport::redraw();
    DrawObject drawer(this);
    for_each(mvpObjects.begin(), mvpObjects.end(), drawer);
    
    glXSwapBuffers(disp, win);
  };
  
  void resize(unsigned int w, unsigned int h) {
    mvWidth=w;
    mvHeight=h;
  };
};

Window openWindow(Display *disp, unsigned int width, unsigned int height) {
  int attrib[] = { GLX_RGBA,
		   GLX_RED_SIZE, 1,
		   GLX_GREEN_SIZE, 1,
		   GLX_BLUE_SIZE, 1,
		   GLX_DEPTH_SIZE, 16,
		   GLX_DOUBLEBUFFER,
		   None };
  int scrnum;
  XSetWindowAttributes attr;
  unsigned long mask;
  Window root;
  Window win;
  GLXContext ctx;
  XVisualInfo *visinfo;
  
  scrnum = DefaultScreen( disp );
  root = RootWindow( disp, scrnum );
  
  visinfo = glXChooseVisual( disp, scrnum, attrib );
  if (!visinfo) throw FeError("Error: couldn't get appropriate visual\n");
  
  /* window attributes */
  attr.background_pixel = 0;
  attr.border_pixel = 0;
  attr.colormap = XCreateColormap( disp, root, visinfo->visual, AllocNone);
  attr.event_mask = StructureNotifyMask | ExposureMask | ButtonPressMask | ButtonReleaseMask | ButtonMotionMask | KeyPressMask | KeyReleaseMask | PointerMotionMask;
  mask = CWBackPixel | CWBorderPixel | CWColormap | CWEventMask;
  
  win = XCreateWindow( disp, root, 0, 0, width, height,
		       0, visinfo->depth, InputOutput,
		       visinfo->visual, mask, &attr );
  XStoreName(disp, win, "xfailure");
  
  ctx = glXCreateContext( disp, visinfo, NULL, True );
  
  glXMakeCurrent( disp, win, ctx );
  
  return win;
}

void displayModel(FeGLModel *model) {
  cerr << endl;
  cerr << "------------------" << endl;
  cerr << " Displaying model" << endl;
  cerr << "------------------" << endl << endl;
  cerr << "  Use middle button or Ctrl+t to translate," << endl;
  cerr << "      Ctrl+middle button or Crtl+r to rotate and" << endl;
  cerr << "      Shift+middle button or Ctrl+s to scale view" << endl;
  cerr << "      'w' for wireframe display" << endl ;
  cerr << "      'h' for hiddenline display" << endl;
  cerr << "      's' for shaded display" << endl;
  cerr << "      'n' for normals toggle" << endl;
  //  cerr << "      't' for top view" << endl;
  //  cerr << "      'f' for front view" << endl;
  //  cerr << "      'r' for right view" << endl;
  cerr << "      'a' for axonometry view" << endl;
  cerr << "      'p' for perspective view" << endl;
  cerr << "      'q' for quit" << endl << endl;  
  
  FeViewport view(0, vpAxo, vpHiddenline);
  view.addObject(model);
  
  Display *disp=XOpenDisplay(NULL);
  Window win=openWindow(disp, 640, 480);
  XMapWindow(disp, win);
  
  double oldMouseX, oldMouseY;
  
  bool endProg=false;
  double dx, dy;
  XEvent ev;
  do {
    XNextEvent(disp, &ev);
    viewportMotionMode mmode;
    
    switch (ev.type) {
      case Expose:
	view.redraw(disp, ev.xany.window);
	break;
      case ConfigureNotify:
	view.resize(ev.xconfigure.width, ev.xconfigure.height);
	view.redraw(disp, ev.xany.window);
	break;
      case MotionNotify:
	dx= ev.xmotion.x - oldMouseX;
	dy=-ev.xmotion.y + oldMouseY;
	oldMouseX=ev.xmotion.x;
	oldMouseY=ev.xmotion.y;

	mmode=view.getMotionMode();

	if (mmode != vpNoMotion) {
	  switch (mmode) {
	    case vpTranslate: view.camTranslate(dx, dy); break;
	    case vpRotate: view.camRotateCenter(dx, dy); break;
	    case vpScale: view.camScale(dx, dy); break;
	  }

	  view.redraw(disp, win);
	} else {
	  if (ev.xmotion.state & Button2Mask) {
	    if (ev.xmotion.state & ControlMask) {      
	      if (ev.xmotion.state & ShiftMask) view.camRotateCenter(dx, dy);
	      else view.camRotateEye(dx, dy);
	    } else if (ev.xmotion.state & ShiftMask) view.camScale(dx, dy);
	    else view.camTranslate(dx, dy);
	    
	    view.redraw(disp, win);
	  }
	}

	break;
      case ButtonPress:
	oldMouseX=ev.xbutton.x;
	oldMouseY=ev.xbutton.y;
	if (ev.xbutton.button == 2) {
	  view.drawCenterFlag=true;
	  view.redraw(disp, win);
	}
	break;
      case ButtonRelease:
	if (ev.xbutton.button == 2) {
	  view.drawCenterFlag=false;
	  view.redraw(disp, win);
	}
	break;
      case KeyPress:
	if (ev.xkey.state & ControlMask) {
	  bool startMotion=false;

	  switch (XKeycodeToKeysym(disp, ev.xkey.keycode, 0)) {
	    case XK_t: case XK_T: 
	      if (view.getMotionMode() != vpTranslate) {
		view.setMotionMode(vpTranslate);
		startMotion=true;
	      }
	      break;
	    case XK_r: case XK_R:
	      if (view.getMotionMode() != vpRotate) {
		view.setMotionMode(vpRotate);
		startMotion=true;
	      }
	      break;
	    case XK_s: case XK_S:
	      if (view.getMotionMode() != vpScale) {
		view.setMotionMode(vpScale);
		startMotion=true;
	      }
	      break;
	  }

	  if (startMotion & view.getMotionMode() == vpNoMotion) {
	    oldMouseX=ev.xkey.x;
	    oldMouseY=ev.xkey.y;
	    view.drawCenterFlag=false;
	    view.redraw(disp, win);
	    cout << "start motion" << endl;
	  }
	} else {
	  switch (XKeycodeToKeysym(disp, ev.xkey.keycode, 0)) {
  	    case XK_Q:
	    case XK_q:
	      endProg=true;
	      break;
	    case XK_W:
 	    case XK_w:
	      view.setRenderMode(vpWireframe);
	      view.redraw(disp, win);
	      break;
 	    case XK_H:
	    case XK_h:
	      view.setRenderMode(vpHiddenline);
	      view.redraw(disp, win);
	      break;
	    case XK_S:
	    case XK_s:
	      view.setRenderMode(vpFlat);
	      view.redraw(disp, win);
	      break;
	    case XK_N:
	    case XK_n:
	      view.drawNormalsFlag=!view.drawNormalsFlag;
	      view.redraw(disp, win);	      
	      break;
	      /* 	    case XK_T:
	    case XK_t:
	      view.setType(vpTop);
	      view.redraw(disp, win);
	      break;
 	    case XK_F:
	    case XK_f:
	      view.setType(vpFront);
	      view.redraw(disp, win);
	      break;
	    case XK_R:
	    case XK_r:
	      view.setType(vpRight);
	      view.redraw(disp, win);
	      break;*/
 	    case XK_A:
	    case XK_a:
	      view.setType(vpAxo);
	      view.redraw(disp, win);
	      break;
	    case XK_P:
  	    case XK_p:
	      view.setType(vpPersp);
	      view.redraw(disp, win);
	      break;
	  }
	}
	break;
      case KeyRelease:
	if (view.getMotionMode() != vpNoMotion) {
	  view.setMotionMode(vpNoMotion);
	  view.drawCenterFlag=false;
	  view.redraw(disp, win);
	}
	break;
    }
  } while(!endProg);
}

int main(int argc, char *argv[]) {
  try {
    printInfo();

    processParams(argc, argv);

    DEBUG(9, cout << "Reading input" << endl);
    
    FeGLModel model;
    FeParser parser(&model);

    DEBUG(2, model.info(cout));
       
    DEBUG(9, cout << "Solving" << endl);

    doCalculations(model);

    DEBUG(9, cout << "Displaying" << endl);

    displayModel(&model);

    DEBUG(9, cout << "Ending program" << endl);
  }
  catch (FeException exc) { cout << "Aborted. " << exc << endl; }
}

/*--- End of file ---*/
