
#ifndef _VIEWPORT_H
#define _VIEWPORT_H

#include <GL/gl.h>

#include "libOGL/OGL.h"

class Viewport;            // Generic viewport
class GLViewport;          // Viewport using OpenGL
class GLDrawable;          // Generic class that can be displayed in GL viewport
class ViewportManager;     // Does opening/closing of viewports, arranging ...

// Default viewport size
#define ViewWidth  192
#define ViewHeight 192

// Default viewport background
#define ViewBgRed   0.00
#define ViewBgGreen 0.33
#define ViewBgBlue  0.39

typedef enum {vpTop, vpFront, vpRight, vpAxo, vpPersp} viewportType;
typedef enum {vpWireframe, vpHiddenline, vpFlat, vpSmooth} viewportRenderMode;
typedef enum {vpNoMotion, vpTranslate, vpRotate, vpScale} viewportMotionMode;

class Viewport {

protected:

  int number;                         // Number of viewport
  int rcX, rcY, rcW, rcH;             // Rendering area

public:

  // Rendering related
  viewportType type;                  // Type of viewport
  viewportRenderMode mode;            // Rendering mode
  viewportMotionMode motionMode;

  bool drawGridFlag;
  bool drawCenterFlag;
  bool drawCamerasFlag;
  bool drawLightsFlag;
  bool drawObjectsFlag;
  bool drawBoundsFlag;
  bool drawNormalsFlag;

  // Camera related
  OGLVector3 eye, center, up;
  OGLFloat fov;
  OGLFloat scale;
  GLdouble zInvert;

public:
  Viewport(int, viewportType, viewportRenderMode);
  virtual ~Viewport();

  // Viewport attributes management
  virtual void setType(viewportType);              // Switches viewport type
  virtual void setRenderMode(viewportRenderMode);  // Switches viewport render mode
  virtual void setMotionMode(viewportMotionMode);
  virtual viewportMotionMode getMotionMode();
  void setRenderingArea(unsigned int x, unsigned int y, unsigned int w, unsigned int h);

  // Camera related
  virtual void defaultView();
  virtual void camRotateCenter(double, double);
  virtual void camRotateEye(double, double);
  virtual void camScale(double, double);
  virtual void camTranslate(double, double);

  // Default drawing routine
  virtual void redraw();
  
  // Routines, that should (or can in some cases) be implemented by child viewports
  virtual void startRender() {};           // Called before start of rendering
  virtual void endRender() {};             // Called after rendering
  virtual void drawGrid()=0;               // Draws grid
  virtual void drawCenter()=0;             // Draws camera target
  virtual void drawCameras()=0;            // Should draw cameras
  virtual void drawLights()=0;             // Should draw lights (not perform lighting)
  virtual void drawObjects()=0;            // Should draw objects
  virtual void drawBounds()=0;             // Should draw objects bounds (mostly boxes, but you can draw anything)
  virtual void drawNormals()=0;            // Should draw normal vectors of all objects
};

class GLViewport: public Viewport {

protected:

  GLclampf bgColor[3];                 // Actual background color
  GLdouble modelviewMatrix[16];        // Actual modelview matrix (for gluProject and gluUnproject)
  GLdouble projectionMatrix[16];       // Actual projection matix (for gluProject and gluUnproject)

  GLuint boundsDisplayList;            // Display list used for all bounds
  bool invalidateDisplayLists;         // Set after rendering mode changes

  void setLighting();

public:

  GLViewport(int, viewportType, viewportRenderMode);

  virtual void redraw();
  virtual void startRender();
  virtual void drawGrid();
  virtual void drawCenter();
  virtual void drawCameras();
  virtual void drawLights();
  virtual void drawObjects();
  virtual void drawBounds();
  virtual void drawNormals();

  virtual void setRenderMode(viewportRenderMode);  // Switches viewport render mode
};


class GLDrawable {
private:
  GLuint displayList;

public:
  GLDrawable();
  virtual ~GLDrawable();

  void invalidateDisplayList();
  void drawToGLViewport(GLViewport *, bool invalidateList=false);

  // This must implement child class
  virtual void glTransformation()=0;          // Sets tranformation for drawing (using GL calls)
  virtual void glDraw(GLViewport *)=0;        // Draw object using GL calls
  virtual void glDrawBound(GLViewport *)=0;   // Draw bounds of object (usualy box, but you can draw anything)
  virtual void glDrawNormals(GLViewport *)=0; // Draw normals of object
};

#endif
